@@ -23,7 +23,17 @@ class CollectionModel: MLBaseModel {
 
     var files = [VLCMLMedia]()
 
-    var cellType: BaseCollectionViewCell.Type { return MediaCollectionViewCell.self }
+    var cellType: BaseCollectionViewCell.Type {
+        if mediaCollection is VLCMLMediaGroup {
+            return UserDefaults.standard.bool(forKey: "\(kVLCVideoLibraryGridLayout)\(String(describing: type(of: mediaCollection)) + name)") ?
+                                              MovieCollectionViewCell.self : MediaCollectionViewCell.self
+        } else {
+            return UserDefaults.standard.bool(forKey: "\(kVLCAudioLibraryGridLayout)\(String(describing: type(of: mediaCollection)) + name)") ?
+                                              MediaGridCollectionCell.self : MediaCollectionViewCell.self
+        }
+    }
+
+    var name: String = "Collections"
 
     // No indicator for this model so no localization needed.
     var indicatorName: String = "Collections"
@@ -65,14 +75,11 @@ class CollectionModel: MLBaseModel {
             do {
                 for case let media in items {
                     if let mainFile = media.mainFile() {
-                        try FileManager.default.removeItem(atPath: mainFile.mrl.path)
+                        mainFile.delete()
                     }
                 }
                 medialibrary.reload()
             }
-            catch let error as NSError {
-                assertionFailure("CollectionModel: Delete failed: \(error.localizedDescription)")
-            }
             filterFilesFromDeletion(of: items)
         }
     }
@@ -113,6 +120,18 @@ extension CollectionModel: MediaLibraryObserver {
         }
     }
 
+    func medialibrary(_ medialibrary: MediaLibraryService,
+                      thumbnailReady media: VLCMLMedia,
+                      type: VLCMLThumbnailSizeType, success: Bool) {
+        guard success else {
+            return
+        }
+        files = mediaCollection.files() ?? []
+        observable.observers.forEach() {
+            $0.value.observer?.mediaLibraryBaseModelReloadView()
+        }
+    }
+
     func medialibraryDidStartRescan() {
         files.removeAll()
     }
