@@ -2,14 +2,18 @@
  * MediaLibraryService.swift
  * VLC for iOS
  *****************************************************************************
- * Copyright © 2018 VideoLAN. All rights reserved.
- * Copyright © 2018 Videolabs
+ * Copyright © 2018-2022 VideoLAN. All rights reserved.
+ * Copyright © 2018-2022 Videolabs
  *
  * Authors: Soomin Lee <bubu # mikan.io>
  *
  * Refer to the COPYING file of the official project for license.
  *****************************************************************************/
 
+import CoreSpotlight
+import VLCMediaLibraryKit
+import UIKit
+
 // MARK: - Notification names
 
 extension Notification.Name {
@@ -23,7 +27,7 @@ extension NSNotification {
 
 // MARK: -
 
-@objc protocol MediaLibraryObserver: class {
+@objc protocol MediaLibraryObserver: AnyObject {
     // Video
     @objc optional func medialibrary(_ medialibrary: MediaLibraryService,
                                      didAddVideos videos: [VLCMLMedia])
@@ -108,53 +112,58 @@ extension NSNotification {
     @objc optional func medialibraryDidStartRescan()
 }
 
-// MARK: -
-
-protocol MediaLibraryMigrationDelegate: class {
-    func medialibraryDidStartMigration(_ medialibrary: MediaLibraryService)
-
-    func medialibraryDidFinishMigration(_ medialibrary: MediaLibraryService)
-
-    func medialibraryDidStopMigration(_ medialibrary: MediaLibraryService)
-}
-
 // MARK: - Delegate for "Backup Media Library" setting
 
-@objc protocol MediaLibraryDeviceBackupDelegate: class {
+@objc protocol MediaLibraryDeviceBackupDelegate: AnyObject {
     @objc func medialibraryDidStartExclusion()
 
     @objc func medialibraryDidCompleteExclusion()
 }
 
+// MARK: - Delegate for hiding Media Library
+
+@objc protocol MediaLibraryHidingDelegate: AnyObject {
+    @objc func medialibraryDidStartHiding()
+
+    @objc func medialibraryDidCompleteHiding()
+}
+
 // MARK: -
 
 class MediaLibraryService: NSObject {
     private static let databaseName: String = "medialibrary.db"
-    private static let migrationKey: String = "MigratedToVLCMediaLibraryKit"
     private static let didForceRescan: String = "MediaLibraryDidForceRescan"
 
-    private var didMigrate = UserDefaults.standard.bool(forKey: MediaLibraryService.migrationKey)
     private var didFinishDiscovery = false
-    // Using ObjectIdentifier to avoid duplication and facilitate
-    // identification of observing object
-    private var observers = [ObjectIdentifier: Observer]()
+
+    private var desiredThumbnailWidth = UInt(320)
+    private var desiredThumbnailHeight = UInt(200)
+
+    private(set) var observable = Observable<MediaLibraryObserver>()
 
     private(set) lazy var medialib = VLCMediaLibrary()
 
-    weak var migrationDelegate: MediaLibraryMigrationDelegate?
     @objc weak var deviceBackupDelegate: MediaLibraryDeviceBackupDelegate?
+    @objc weak var hidingDelegate: MediaLibraryHidingDelegate?
 
     @objc var isExcludingFromBackup: Bool = false
+    @objc var isHidingLibrary: Bool = false
 
     override init() {
         super.init()
-        medialib.delegate = self
         setupMediaLibrary()
+        medialib.delegate = self
         NotificationCenter.default.addObserver(self, selector: #selector(reload),
                                                name: .VLCNewFileAddedNotification, object: nil)
 
         NotificationCenter.default.addObserver(self, selector: #selector(handleWillEnterForegroundNotification),
                                                name: UIApplication.willEnterForegroundNotification, object: nil)
+
+        let screen = UIScreen.main
+        let cellSizeWidth = MovieCollectionViewCell.cellSizeForWidth(screen.bounds.width).width
+        let scaledCellWidth = cellSizeWidth * screen.scale
+        desiredThumbnailWidth = UInt(scaledCellWidth)
+        desiredThumbnailHeight = UInt(scaledCellWidth / 1.6)
     }
 }
 
@@ -170,13 +179,18 @@ private extension MediaLibraryService {
 
     private func startMediaLibrary(on path: String) {
         let excludeMediaLibrary = !UserDefaults.standard.bool(forKey: kVLCSettingBackupMediaLibrary)
+        let hideML = UserDefaults.standard.bool(forKey: kVLCSettingHideLibraryInFilesApp)
         excludeFromDeviceBackup(excludeMediaLibrary)
+        hideMediaLibrary(hideML)
 
         if UserDefaults.standard.bool(forKey: MediaLibraryService.didForceRescan) == false {
             medialib.forceRescan()
             UserDefaults.standard.set(true, forKey: MediaLibraryService.didForceRescan)
         }
 
+        FileManager.default.createFile(atPath: "\(path)/\(NSLocalizedString("MEDIALIBRARY_FILES_PLACEHOLDER", comment: ""))", contents: nil, attributes: nil)
+        try? FileManager.default.removeItem(atPath: "\(path)/\(NSLocalizedString("MEDIALIBRARY_ADDING_PLACEHOLDER", comment: ""))")
+
         medialib.reload()
         medialib.discover(onEntryPoint: "file://" + path)
     }
@@ -221,140 +235,6 @@ private extension MediaLibraryService {
     }
 }
 
-// MARK: - Migration
-
-private extension MediaLibraryService {
-    func startMigrationIfNeeded() {
-        guard !didMigrate else {
-            return
-        }
-        migrationDelegate?.medialibraryDidStartMigration(self)
-
-        migrateToNewMediaLibrary() {
-            [unowned self] success in
-            if success {
-                self.migrationDelegate?.medialibraryDidFinishMigration(self)
-            } else {
-                self.migrationDelegate?.medialibraryDidStopMigration(self)
-            }
-        }
-    }
-
-    func migrateMedia(_ oldMedialibrary: MLMediaLibrary,
-                      completionHandler: @escaping (Bool) -> Void) {
-        guard let allFiles = MLFile.allFiles() as? [MLFile] else {
-            assertionFailure("MediaLibraryService: Migration: Unable to retrieve all files")
-            completionHandler(false)
-            return
-        }
-
-        for media in allFiles {
-            if let newMedia = fetchMedia(with: media.url) {
-                newMedia.updateTitle(media.title)
-                newMedia.setPlayCount(media.playCount.uint32Value)
-                newMedia.setMetadataOf(.progress, intValue: media.lastPosition.int64Value)
-                newMedia.setMetadataOf(.seen, intValue: media.unread.int64Value)
-                // Only delete files that are not in playlist
-                if media.labels.isEmpty {
-                    oldMedialibrary.remove(media)
-                }
-            }
-        }
-
-        oldMedialibrary.save {
-            success in
-            completionHandler(success)
-        }
-    }
-
-    // This private method migrates old playlist and removes file and playlist
-    // from the old medialibrary.
-    // Note: This removes **only** files that are in a playlist
-    func migratePlaylists(_ oldMedialibrary: MLMediaLibrary,
-                          completionHandler: @escaping (Bool) -> Void) {
-        guard let allLabels = MLLabel.allLabels() as? [MLLabel] else {
-            assertionFailure("MediaLibraryService: Migration: Unable to retrieve all labels")
-            completionHandler(false)
-            return
-        }
-
-        for label in allLabels {
-            guard let newPlaylist = createPlaylist(with: label.name) else {
-                assertionFailure("MediaLibraryService: Migration: Unable to create playlist.")
-                continue
-            }
-
-            guard let files = label.files as? Set<MLFile> else {
-                assertionFailure("MediaLibraryService: Migration: Unable to retrieve files from label")
-                oldMedialibrary.remove(label)
-                continue
-            }
-
-            for file in files {
-                if let newMedia = fetchMedia(with: file.url) {
-                    if newPlaylist.appendMedia(withIdentifier: newMedia.identifier()) {
-                        oldMedialibrary.remove(file)
-                    }
-                }
-            }
-            oldMedialibrary.remove(label)
-        }
-        oldMedialibrary.save() {
-            success in
-            completionHandler(success)
-        }
-    }
-
-    func migrateToNewMediaLibrary(completionHandler: @escaping (Bool) -> Void) {
-        guard let oldMedialibrary = MLMediaLibrary.sharedMediaLibrary() as? MLMediaLibrary else {
-            assertionFailure("MediaLibraryService: Migration: Unable to retrieve old medialibrary")
-            completionHandler(false)
-            return
-        }
-
-        migrateMedia(oldMedialibrary) {
-            [unowned self] success in
-
-            guard success else {
-                assertionFailure("MediaLibraryService: Failed to migrate Media.")
-                completionHandler(false)
-                return
-            }
-
-            self.migratePlaylists(oldMedialibrary) {
-                [unowned self] success in
-                if success {
-                    UserDefaults.standard.set(true, forKey: MediaLibraryService.migrationKey)
-                    self.didMigrate = true
-                } else {
-                    assertionFailure("MediaLibraryService: Failed to migrate Playlist.")
-                }
-                completionHandler(success)
-            }
-        }
-    }
-}
-
-// MARK: - Observer
-
-private extension MediaLibraryService {
-    struct Observer {
-        weak var observer: MediaLibraryObserver?
-    }
-}
-
-extension MediaLibraryService {
-    func addObserver(_ observer: MediaLibraryObserver) {
-        let identifier = ObjectIdentifier(observer)
-        observers[identifier] = Observer(observer: observer)
-    }
-
-    func removeObserver(_ observer: MediaLibraryObserver) {
-        let identifier = ObjectIdentifier(observer)
-        observers.removeValue(forKey: identifier)
-    }
-}
-
 // MARK: - Helpers
 
 @objc extension MediaLibraryService {
@@ -412,8 +292,14 @@ extension MediaLibraryService {
         mlMedia.titleIndex = Int64(player.indexOfCurrentTitle)
 
         if mlMedia.type() == .video {
-            mlMedia.requestThumbnail(of: .thumbnail, desiredWidth: 320,
-                                     desiredHeight: 200, atPosition: player.playbackPosition)
+            if let thumbnailURL = mlMedia.thumbnail() {
+                if mlMedia.progress < 0.90 {
+                    _ = try? FileManager.default.removeItem(at: thumbnailURL)
+                    VLCThumbnailsCache.invalidateThumbnail(for: thumbnailURL)
+                }
+            }
+            mlMedia.requestThumbnail(of: .thumbnail, desiredWidth: desiredThumbnailWidth,
+                                     desiredHeight: desiredThumbnailHeight, atPosition: player.playbackPosition)
         }
     }
 
@@ -436,6 +322,40 @@ extension MediaLibraryService {
             }
         }
     }
+
+    @objc func hideMediaLibrary(_ hide: Bool) {
+        if let documentPath = NSSearchPathForDirectoriesInDomains(.documentDirectory, .userDomainMask, true).first {
+            var documentURL = URL(fileURLWithPath: documentPath)
+            isHidingLibrary = true
+            hidingDelegate?.medialibraryDidStartHiding()
+            DispatchQueue.global().async {
+                documentURL.setHidden(hide, recursive: true, onlyFirstLevel: false) {
+                    self.isHidingLibrary = false
+                    self.hidingDelegate?.medialibraryDidCompleteHiding()
+                }
+            }
+        }
+    }
+
+    @objc func exportMediaLibrary() {
+        guard let documentPath = NSSearchPathForDirectoriesInDomains(.documentDirectory, .userDomainMask, true).first,
+            let libraryPath = NSSearchPathForDirectoriesInDomains(.libraryDirectory, .userDomainMask, true).first else {
+                preconditionFailure("MediaLibraryService: Unable to find medialibrary.")
+        }
+
+        let databasePath = libraryPath + "/MediaLibrary/" + MediaLibraryService.databaseName
+        let targetPath = documentPath + "/Logs/" + MediaLibraryService.databaseName
+
+        do {
+            try FileManager.default.createDirectory(atPath: targetPath,
+                                                    withIntermediateDirectories: true)
+        } catch let error as NSError {
+            assertionFailure("Failed to create directory: \(error.localizedDescription)")
+        }
+
+        _ = try? FileManager.default.removeItem(atPath: targetPath)
+        _ = try? FileManager.default.copyItem(atPath: databasePath, toPath: targetPath)
+    }
 }
 
 // MARK: - Application notifications
@@ -450,10 +370,15 @@ extension MediaLibraryService {
         // for the user that can be created by deleting media from the Files app.
         // This could lead to disk space issues.
         // For now since we do not handle restoration, we delete the `.Trash` folder every time.
-         _ = try? FileManager.default.removeItem(atPath: documentPath + "/.Trash")
+        // Run this on a background queue to not block the main thread and have the app killed on launch for taking too long
+        DispatchQueue.global(qos: .userInitiated).async {
+            _ = try? FileManager.default.removeItem(atPath: documentPath + "/.Trash")
 
-        // Reload in order to make sure that there is no old artifacts left
-        reload()
+            DispatchQueue.main.async {
+                // Reload in order to make sure that there is no old artifacts left
+                self.reload()
+            }
+        }
     }
 }
 
@@ -461,7 +386,7 @@ extension MediaLibraryService {
 
 @objc extension MediaLibraryService {
     func artists(sortingCriteria sort: VLCMLSortingCriteria = .alpha,
-                 desc: Bool = false, listAll all: Bool = false) -> [VLCMLArtist] {
+                 desc: Bool = false, listAll all: Bool = true) -> [VLCMLArtist] {
         return medialib.artists(with: sort, desc: desc, all: all) ?? []
     }
 
@@ -484,7 +409,7 @@ extension MediaLibraryService {
             assertionFailure("MediaLibraryService: requestThumbnail: unknown case.")
         }
 
-        if !media.requestThumbnail(of: .thumbnail, desiredWidth: 320, desiredHeight: 200, atPosition: 0.03) {
+        if !media.requestThumbnail(of: .thumbnail, desiredWidth: desiredThumbnailWidth, desiredHeight: desiredThumbnailHeight, atPosition: 0.03) {
             assertionFailure("MediaLibraryService: Failed to generate thumbnail for: \(media.identifier())")
         }
     }
@@ -547,7 +472,7 @@ extension MediaLibraryService: VLCMediaLibraryDelegate {
         let videos = media.filter {( $0.type() == .video )}
         let tracks = media.filter {( $0.type() == .audio )}
 
-        for observer in observers {
+        for observer in observable.observers {
             observer.value.observer?.medialibrary?(self, didAddVideos: videos)
             observer.value.observer?.medialibrary?(self, didAddTracks: tracks)
         }
@@ -571,7 +496,7 @@ extension MediaLibraryService: VLCMediaLibraryDelegate {
         let tracks = media.filter {( $0.type() == .audio)}
 
         // Shows and albumtracks are known only after when the medialibrary calls didModifyMedia
-        for observer in observers {
+        for observer in observable.observers {
             observer.value.observer?.medialibrary?(self, didAddShowEpisodes: showEpisodes)
             observer.value.observer?.medialibrary?(self, didAddAlbumTracks: albumTrack)
             observer.value.observer?.medialibrary?(self, didModifyVideos: videos)
@@ -584,14 +509,14 @@ extension MediaLibraryService: VLCMediaLibraryDelegate {
         mediaIds.forEach { stringIds.append("\($0)") }
         CSSearchableIndex.default().deleteSearchableItems(withIdentifiers: stringIds, completionHandler: nil)
 
-        for observer in observers {
+        for observer in observable.observers {
             observer.value.observer?.medialibrary?(self, didDeleteMediaWithIds: mediaIds)
         }
     }
 
     func medialibrary(_ medialibrary: VLCMediaLibrary, thumbnailReadyFor media: VLCMLMedia,
                       of type: VLCMLThumbnailSizeType, withSuccess success: Bool) {
-        for observer in observers {
+        for observer in observable.observers {
             observer.value.observer?.medialibrary?(self, thumbnailReady: media,
                                                    type: type, success: success)
         }
@@ -602,20 +527,20 @@ extension MediaLibraryService: VLCMediaLibraryDelegate {
 
 extension MediaLibraryService {
     func medialibrary(_ medialibrary: VLCMediaLibrary, didAdd artists: [VLCMLArtist]) {
-        for observer in observers {
+        for observer in observable.observers {
             observer.value.observer?.medialibrary?(self, didAddArtists: artists)
         }
     }
 
     func medialibrary(_ medialibrary: VLCMediaLibrary,
                       didModifyArtistsWithIds artistsIds: [NSNumber]) {
-        for observer in observers {
+        for observer in observable.observers {
             observer.value.observer?.medialibrary?(self, didModifyArtistsWithIds: artistsIds)
         }
     }
 
     func medialibrary(_ medialibrary: VLCMediaLibrary, didDeleteArtistsWithIds artistsIds: [NSNumber]) {
-        for observer in observers {
+        for observer in observable.observers {
             observer.value.observer?.medialibrary?(self, didDeleteArtistsWithIds: artistsIds)
         }
     }
@@ -625,20 +550,20 @@ extension MediaLibraryService {
 
 extension MediaLibraryService {
     func medialibrary(_ medialibrary: VLCMediaLibrary, didAdd albums: [VLCMLAlbum]) {
-        for observer in observers {
+        for observer in observable.observers {
             observer.value.observer?.medialibrary?(self, didAddAlbums: albums)
         }
     }
 
     func medialibrary(_ medialibrary: VLCMediaLibrary,
                       didModifyAlbumsWithIds albumsIds: [NSNumber]) {
-        for observer in observers {
+        for observer in observable.observers {
             observer.value.observer?.medialibrary?(self, didModifyAlbumsWithIds: albumsIds)
         }
     }
 
     func medialibrary(_ medialibrary: VLCMediaLibrary, didDeleteAlbumsWithIds albumsIds: [NSNumber]) {
-        for observer in observers {
+        for observer in observable.observers {
             observer.value.observer?.medialibrary?(self, didDeleteAlbumsWithIds: albumsIds)
         }
     }
@@ -648,21 +573,21 @@ extension MediaLibraryService {
 
 extension MediaLibraryService {
     func medialibrary(_ medialibrary: VLCMediaLibrary, didAdd genres: [VLCMLGenre]) {
-        for observer in observers {
+        for observer in observable.observers {
             observer.value.observer?.medialibrary?(self, didAddGenres: genres)
         }
     }
 
     func medialibrary(_ medialibrary: VLCMediaLibrary,
                       didModifyGenresWithIds genresIds: [NSNumber]) {
-        for observer in observers {
+        for observer in observable.observers {
             observer.value.observer?.medialibrary?(self, didModifyGenresWithIds: genresIds)
         }
     }
 
     func medialibrary(_ medialibrary: VLCMediaLibrary,
                       didDeleteGenresWithIds genresIds: [NSNumber]) {
-        for observer in observers {
+        for observer in observable.observers {
             observer.value.observer?.medialibrary?(self, didDeleteGenresWithIds: genresIds)
         }
     }
@@ -672,20 +597,20 @@ extension MediaLibraryService {
 
 extension MediaLibraryService {
     func medialibrary(_ medialibrary: VLCMediaLibrary, didAdd playlists: [VLCMLPlaylist]) {
-        for observer in observers {
+        for observer in observable.observers {
             observer.value.observer?.medialibrary?(self, didAddPlaylists: playlists)
         }
     }
 
     func medialibrary(_ medialibrary: VLCMediaLibrary,
                       didModifyPlaylistsWithIds playlistsIds: [NSNumber]) {
-        for observer in observers {
+        for observer in observable.observers {
             observer.value.observer?.medialibrary?(self, didModifyPlaylistsWithIds: playlistsIds)
         }
     }
 
     func medialibrary(_ medialibrary: VLCMediaLibrary, didDeletePlaylistsWithIds playlistsIds: [NSNumber]) {
-        for observer in observers {
+        for observer in observable.observers {
             observer.value.observer?.medialibrary?(self, didDeletePlaylistsWithIds: playlistsIds)
         }
     }
@@ -695,7 +620,7 @@ extension MediaLibraryService {
 
 extension MediaLibraryService {
     func medialibrary(_ medialibrary: VLCMediaLibrary, didAdd mediaGroups: [VLCMLMediaGroup]) {
-        for observer in observers {
+        for observer in observable.observers {
             observer.value.observer?.medialibrary?(self,
                                                    didAddMediaGroups: mediaGroups)
         }
@@ -703,7 +628,7 @@ extension MediaLibraryService {
 
     func medialibrary(_ medialibrary: VLCMediaLibrary,
                       didModifyMediaGroupsWithIds mediaGroupsIds: [NSNumber]) {
-        for observer in observers {
+        for observer in observable.observers {
             observer.value.observer?.medialibrary?(self,
                                                    didModifyMediaGroupsWithIds: mediaGroupsIds)
         }
@@ -711,7 +636,7 @@ extension MediaLibraryService {
 
     func medialibrary(_ medialibrary: VLCMediaLibrary,
                       didDeleteMediaGroupsWithIds mediaGroupsIds: [NSNumber]) {
-        for observer in observers {
+        for observer in observable.observers {
             observer.value.observer?.medialibrary?(self,
                                                    didDeleteMediaGroupsWithIds: mediaGroupsIds)
         }
@@ -732,9 +657,6 @@ extension MediaLibraryService {
     }
 
     func medialibrary(_ medialibrary: VLCMediaLibrary, didUpdateParsingStatsWithPercent percent: UInt32) {
-        if didFinishDiscovery && percent == 100 {
-             startMigrationIfNeeded()
-        }
     }
 }
 
@@ -757,7 +679,7 @@ extension MediaLibraryService {
 
 extension MediaLibraryService {
     func medialibraryDidStartRescan(_ medialibrary: VLCMediaLibrary) {
-        for observer in observers {
+        for observer in observable.observers {
             observer.value.observer?.medialibraryDidStartRescan?()
         }
     }
