@@ -2,7 +2,7 @@
  * VLCPlaybackService.m
  * VLC for iOS
  *****************************************************************************
- * Copyright (c) 2013-2018 VideoLAN. All rights reserved.
+ * Copyright (c) 2013-2022 VLC authors and VideoLAN
  * $Id$
  *
  * Authors: Felix Paul Kühne <fkuehne # videolan.org>
@@ -12,12 +12,12 @@
  *          Tobias Conradi <videolan # tobias-conradi.de>
  *          Sylver Bruneau <sylver.bruneau # gmail dot com>
  *          Winston Weinert <winston # ml1 dot net>
+ *          Maxime Chapelet <umxprime # videolabs.io>
  *
  * Refer to the COPYING file of the official project for license.
  *****************************************************************************/
 
 #import "VLCPlaybackService.h"
-#import "UIDevice+VLC.h"
 #import <AVFoundation/AVFoundation.h>
 #import "VLCRemoteControlService.h"
 #import "VLCMetadata.h"
@@ -33,7 +33,11 @@ NSString *const VLCPlaybackServicePlaybackMetadataDidChange = @"VLCPlaybackServi
 NSString *const VLCPlaybackServicePlaybackDidFail = @"VLCPlaybackServicePlaybackDidFail";
 NSString *const VLCPlaybackServicePlaybackPositionUpdated = @"VLCPlaybackServicePlaybackPositionUpdated";
 
-@interface VLCPlaybackService () <VLCMediaPlayerDelegate, VLCMediaDelegate, VLCRemoteControlServiceDelegate>
+#if TARGET_OS_IOS
+@interface VLCPlaybackService () <VLCMediaPlayerDelegate, VLCMediaDelegate, VLCMediaListPlayerDelegate, VLCRemoteControlServiceDelegate, EqualizerViewDelegate>
+#else
+@interface VLCPlaybackService () <VLCMediaPlayerDelegate, VLCMediaDelegate, VLCMediaListPlayerDelegate, VLCRemoteControlServiceDelegate>
+#endif
 {
     VLCRemoteControlService *_remoteControlService;
     VLCMediaPlayer *_backgroundDummyPlayer;
@@ -48,7 +52,6 @@ NSString *const VLCPlaybackServicePlaybackPositionUpdated = @"VLCPlaybackService
 
     BOOL _isInFillToScreen;
     NSUInteger _previousAspectRatio;
-    
 
     UIView *_videoOutputViewWrapper;
     UIView *_actualVideoOutputView;
@@ -56,6 +59,7 @@ NSString *const VLCPlaybackServicePlaybackPositionUpdated = @"VLCPlaybackService
 
     BOOL _needsMetadataUpdate;
     BOOL _mediaWasJustStarted;
+    int _majorPositionChangeInProgress;
     BOOL _recheckForExistingThumbnail;
     BOOL _externalAudioPlaybackDeviceConnected;
 
@@ -67,6 +71,8 @@ NSString *const VLCPlaybackServicePlaybackPositionUpdated = @"VLCPlaybackService
     VLCDialogProvider *_dialogProvider;
     VLCCustomDialogRendererHandler *_customDialogHandler;
     VLCPlayerDisplayController *_playerDisplayController;
+
+    NSMutableArray *_openedLocalURLs;
 }
 
 @end
@@ -96,6 +102,7 @@ NSString *const VLCPlaybackServicePlaybackPositionUpdated = @"VLCPlaybackService
 {
     self = [super init];
     if (self) {
+        _fullscreenSessionRequested = YES;
         // listen to audiosessions and appkit callback
         _externalAudioPlaybackDeviceConnected = [self isExternalAudioPlaybackDeviceConnected];
         NSNotificationCenter *defaultCenter = [NSNotificationCenter defaultCenter];
@@ -129,6 +136,10 @@ NSString *const VLCPlaybackServicePlaybackPositionUpdated = @"VLCPlaybackService
         // stay alive in background exclusively for Chromecast.
         _backgroundDummyPlayer = [[VLCMediaPlayer alloc] initWithOptions:@[@"--demux=rawaud"]];
         _backgroundDummyPlayer.media = [[VLCMedia alloc] initWithPath:@"/dev/zero"];
+
+        _mediaList = [[VLCMediaList alloc] init];
+
+        _openedLocalURLs = [[NSMutableArray alloc] init];
     }
     return self;
 }
@@ -162,6 +173,16 @@ NSString *const VLCPlaybackServicePlaybackPositionUpdated = @"VLCPlaybackService
 
     _sessionWillRestart = _playerIsSetup;
     _playerIsSetup ? [self stopPlayback] : [self startPlayback];
+
+#if TARGET_OS_TV
+    VLCFullscreenMovieTVViewController *movieVC = [VLCFullscreenMovieTVViewController fullscreenMovieTVViewController];
+
+    if (![movieVC isBeingPresented]) {
+        [[UIApplication sharedApplication].keyWindow.rootViewController presentViewController:movieVC
+                                                                                     animated:YES
+                                                                                   completion:nil];
+    }
+#endif
 }
 
 - (VLCTime *)playedTime
@@ -198,17 +219,64 @@ NSString *const VLCPlaybackServicePlaybackPositionUpdated = @"VLCPlaybackService
     _actualVideoOutputView.autoresizingMask = UIViewAutoresizingFlexibleWidth | UIViewAutoresizingFlexibleHeight;
     _actualVideoOutputView.autoresizesSubviews = YES;
 
-    _listPlayer = [[VLCMediaListPlayer alloc] initWithDrawable:_actualVideoOutputView];
+    /* the chromecast options cannot be set per media, so we need to set it per
+     * media player instance however, potentially initialising an additional library instance
+     * for this is costly, so this should be done only if needed */
+    NSUserDefaults *userDefaults = [NSUserDefaults standardUserDefaults];
+    BOOL chromecastPassthrough = [[userDefaults objectForKey:kVLCSettingCastingAudioPassthrough] boolValue];
+    int chromecastQuality = [[userDefaults objectForKey:kVLCSettingCastingConversionQuality] intValue];
+    NSMutableArray *libVLCOptions = [NSMutableArray array];
+    if (chromecastPassthrough) {
+        [libVLCOptions addObject:[@"--" stringByAppendingString:kVLCSettingCastingAudioPassthrough]];
+    }
+    if (chromecastQuality != 2) {
+        [libVLCOptions addObject:[NSString stringWithFormat:@"--%@=%i", kVLCSettingCastingConversionQuality, chromecastQuality]];
+    }
+    if (libVLCOptions.count > 0) {
+        _listPlayer = [[VLCMediaListPlayer alloc] initWithOptions:libVLCOptions
+                                                      andDrawable:_actualVideoOutputView];
+    } else {
+        _listPlayer = [[VLCMediaListPlayer alloc] initWithDrawable:_actualVideoOutputView];
+    }
+    _listPlayer.delegate = self;
 
-    /* to enable debug logging for the playback library instance, switch the boolean below
-     * note that the library instance used for playback may not necessarily match the instance
-     * used for media discovery or thumbnailing */
-    _listPlayer.mediaPlayer.libraryInstance.debugLogging = NO;
+#if MEDIA_PLAYBACK_DEBUG
+    _listPlayer.mediaPlayer.libraryInstance.debugLogging = YES;
+    _listPlayer.mediaPlayer.libraryInstance.debugLoggingLevel = 4;
+#endif
+    BOOL saveDebugLogs = [userDefaults boolForKey:kVLCSaveDebugLogs];
+    if (saveDebugLogs) {
+        NSArray *searchPaths;
+#if TARGET_OS_IOS
+        searchPaths = NSSearchPathForDirectoriesInDomains(NSDocumentDirectory, NSUserDomainMask, YES);
+#else
+        searchPaths = NSSearchPathForDirectoriesInDomains(NSCachesDirectory, NSUserDomainMask, YES);
+#endif
+        NSString* logFilePath = [searchPaths[0] stringByAppendingPathComponent:@"Logs"];
+        NSFileManager *fileManager = [NSFileManager defaultManager];
+        ret = [fileManager fileExistsAtPath:logFilePath];
+        if (!ret) {
+            [fileManager createDirectoryAtPath:logFilePath withIntermediateDirectories:YES attributes:nil error:nil];
+        }
+        NSDate *date = [NSDate date];
+        NSDateFormatter *dateFormatter=[[NSDateFormatter alloc] init];
+        [dateFormatter setTimeZone:[NSTimeZone timeZoneWithName:@"UTC"]];
+        [dateFormatter setDateFormat:@"yyyy-MM-dd--HH-mm-ss"];
+        logFilePath = [logFilePath stringByAppendingPathComponent:[NSString stringWithFormat: @"vlcdebug-%@.log", [dateFormatter stringFromDate:date]]];
+        APLog(@"logging at '%@'", logFilePath);
+        [_listPlayer.mediaPlayer.libraryInstance setDebugLoggingToFile:logFilePath];
+    }
 
+    id<VLCFilter> newFilter = _listPlayer.mediaPlayer.adjustFilter;
+    [newFilter applyParametersFrom:_adjustFilter.mediaPlayerAdjustFilter];
+    newFilter.enabled = _adjustFilter.mediaPlayerAdjustFilter.isEnabled;
+    _adjustFilter = [[VLCPlaybackServiceAdjustFilter alloc] initWithMediaPlayerAdjustFilter:newFilter];
     _mediaPlayer = _listPlayer.mediaPlayer;
+
     [_mediaPlayer setDelegate:self];
-    if ([[defaults objectForKey:kVLCSettingPlaybackSpeedDefaultValue] floatValue] != 0)
-        [_mediaPlayer setRate: [[defaults objectForKey:kVLCSettingPlaybackSpeedDefaultValue] floatValue]];
+    CGFloat defaultPlaybackSpeed = [[defaults objectForKey:kVLCSettingPlaybackSpeedDefaultValue] floatValue];
+    if (defaultPlaybackSpeed != 0.)
+        [_mediaPlayer setRate: defaultPlaybackSpeed];
     int deinterlace = [[defaults objectForKey:kVLCSettingDeinterlace] intValue];
     [_mediaPlayer setDeinterlace:deinterlace withFilter:@"blend"];
 
@@ -218,8 +286,12 @@ NSString *const VLCPlaybackServicePlaybackPositionUpdated = @"VLCPlaybackService
     [media addOptions:self.mediaOptionsDictionary];
 
     [_listPlayer setMediaList:self.mediaList];
-
-    [_listPlayer setRepeatMode:VLCDoNotRepeat];
+#if TARGET_OS_IOS
+    if ([defaults boolForKey:kVLCPlayerShouldRememberState]) {
+        VLCRepeatMode repeatMode = [defaults integerForKey:kVLCPlayerIsRepeatEnabled];
+        [_listPlayer setRepeatMode:repeatMode];
+    }
+#endif
 
     [_playbackSessionManagementLock unlock];
 
@@ -248,7 +320,9 @@ NSString *const VLCPlaybackServicePlaybackPositionUpdated = @"VLCPlaybackService
     [_mediaPlayer addObserver:self forKeyPath:@"time" options:0 context:nil];
     [_mediaPlayer addObserver:self forKeyPath:@"remainingTime" options:0 context:nil];
 
+#if !TARGET_OS_TV
     [_mediaPlayer setRendererItem:_renderer];
+#endif
 
     [_listPlayer playItemAtNumber:@(_itemInMediaListToBePlayedFirst)];
 
@@ -303,20 +377,39 @@ NSString *const VLCPlaybackServicePlaybackPositionUpdated = @"VLCPlaybackService
 #endif
             [_mediaPlayer stop];
         }
+
+        if (_playbackCompletion) {
+            BOOL finishedPlaybackWithError = false;
+            if (_mediaPlayer.state == VLCMediaPlayerStateStopped && _mediaPlayer.media != nil) {
+                // Since VLCMediaPlayerStateError is sometimes not matched with a valid media.
+                // This checks for decoded Audio & Video blocks.
+                finishedPlaybackWithError = (_mediaPlayer.media.numberOfDecodedAudioBlocks == 0)
+                                             && (_mediaPlayer.media.numberOfDecodedVideoBlocks == 0);
+            } else {
+                finishedPlaybackWithError = _mediaPlayer.state == VLCMediaPlayerStateError;
+            }
+            finishedPlaybackWithError = finishedPlaybackWithError && !_sessionWillRestart;
+
+            _playbackCompletion(!finishedPlaybackWithError);
+        }
+
         _mediaPlayer = nil;
         _listPlayer = nil;
     }
+
+    for (NSURL *url in _openedLocalURLs) {
+        [url stopAccessingSecurityScopedResource];
+    }
+    _openedLocalURLs = nil;
+    _openedLocalURLs = [[NSMutableArray alloc] init];
+
     if (!_sessionWillRestart) {
         _mediaList = nil;
+        _mediaList = [[VLCMediaList alloc] init];
     }
     _playerIsSetup = NO;
     [_shuffleStack removeAllObjects];
 
-    if (_playbackCompletion) {
-        BOOL finishedPlaybackWithError = _mediaPlayer.state == VLCMediaPlayerStateError &&  !_sessionWillRestart;
-        _playbackCompletion(!finishedPlaybackWithError);
-    }
-
     [[self remoteControlService] unsubscribeFromRemoteCommands];
 
     [_playbackSessionManagementLock unlock];
@@ -328,7 +421,6 @@ NSString *const VLCPlaybackServicePlaybackPositionUpdated = @"VLCPlaybackService
 }
 
 #if TARGET_OS_IOS
-
 - (void)restoreAudioAndSubtitleTrack
 {
     VLCMLMedia *media = [_delegate mediaForPlayingMedia:_mediaPlayer.media];
@@ -361,6 +453,17 @@ NSString *const VLCPlaybackServicePlaybackPositionUpdated = @"VLCPlaybackService
     if ([self.delegate respondsToSelector:@selector(playbackPositionUpdated:)])
         [self.delegate playbackPositionUpdated:self];
 
+    if (_majorPositionChangeInProgress >= 1) {
+        [self.metadata updateExposedTimingFromMediaPlayer:_listPlayer.mediaPlayer];
+        _majorPositionChangeInProgress++;
+
+        /* we wait up to 10 time change intervals for the major position change
+         * to take effect, afterwards we give up, safe battery and let the OS calculate the position */
+        if (_majorPositionChangeInProgress == 10) {
+            _majorPositionChangeInProgress = 0;
+        }
+    }
+
     [[NSNotificationCenter defaultCenter] postNotificationName:VLCPlaybackServicePlaybackPositionUpdated
                                                         object:self];
 }
@@ -397,10 +500,12 @@ NSString *const VLCPlaybackServicePlaybackPositionUpdated = @"VLCPlaybackService
 
 - (BOOL)currentMediaHasTrackToChooseFrom
 {
-    return [[_mediaPlayer audioTrackIndexes] count] > 2 || [[_mediaPlayer videoSubTitlesIndexes] count] > 1;
+    /* allow track selection if there is more than 1 audio track or if there is video because even if
+     * there is only video, there will always be the option to download additional subtitles */
+    return [[_mediaPlayer audioTrackIndexes] count] > 2 || [[_mediaPlayer videoTrackIndexes] count] >= 1;
 }
 
-- (BOOL) isSeekable
+- (BOOL)isSeekable
 {
     return _mediaPlayer.isSeekable;
 }
@@ -439,6 +544,7 @@ NSString *const VLCPlaybackServicePlaybackPositionUpdated = @"VLCPlaybackService
 - (void)setPlaybackPosition:(float)position
 {
     _mediaPlayer.position = position;
+    _majorPositionChangeInProgress = 1;
 }
 
 - (void)setSubtitleDelay:(float)subtitleDeleay
@@ -451,65 +557,6 @@ NSString *const VLCPlaybackServicePlaybackPositionUpdated = @"VLCPlaybackService
     return _mediaPlayer.currentVideoSubTitleDelay/1000.;
 }
 
-- (float)hue
-{
-    return _mediaPlayer.hue;
-}
-
-- (void)setHue:(float)hue
-{
-    _mediaPlayer.hue = hue;
-}
-
-- (float)contrast
-{
-    return _mediaPlayer.contrast;
-}
-
-- (void)setContrast:(float)contrast
-{
-    _mediaPlayer.contrast = contrast;
-}
-
-- (float)brightness
-{
-    return _mediaPlayer.brightness;
-}
-
-- (void)setBrightness:(float)brightness
-{
-    _mediaPlayer.brightness = brightness;
-}
-
-- (float)saturation
-{
-    return _mediaPlayer.saturation;
-}
-
-- (void)setSaturation:(float)saturation
-{
-    _mediaPlayer.saturation = saturation;
-}
-
-- (void)setGamma:(float)gamma
-{
-    _mediaPlayer.gamma = gamma;
-}
-
-- (float)gamma
-{
-    return _mediaPlayer.gamma;
-}
-
-- (void)resetFilters
-{
-    _mediaPlayer.hue = 0.;
-    _mediaPlayer.contrast = 1.;
-    _mediaPlayer.brightness = 1.;
-    _mediaPlayer.saturation = 1.;
-    _mediaPlayer.gamma = 1.;
-}
-
 - (void)toggleRepeatMode
 {
     if (_listPlayer.repeatMode == VLCRepeatAllItems) {
@@ -541,12 +588,12 @@ NSString *const VLCPlaybackServicePlaybackPositionUpdated = @"VLCPlaybackService
 
 - (NSInteger)numberOfAudioTracks
 {
-    return _mediaPlayer.audioTrackIndexes.count;
+    return [_mediaPlayer numberOfAudioTracks];
 }
 
 - (NSInteger)numberOfVideoSubtitlesIndexes
 {
-    return _mediaPlayer.videoSubTitlesIndexes.count;
+    return _mediaPlayer.videoSubTitlesIndexes.count + 1;
 }
 
 - (NSInteger)numberOfTitles
@@ -561,23 +608,27 @@ NSString *const VLCPlaybackServicePlaybackPositionUpdated = @"VLCPlaybackService
 
 - (NSString *)videoSubtitleNameAtIndex:(NSInteger)index
 {
-    if (index >= 0 && index < _mediaPlayer.videoSubTitlesNames.count)
+    NSInteger count = _mediaPlayer.videoSubTitlesNames.count;
+    if (index >= 0 && index < count) {
         return _mediaPlayer.videoSubTitlesNames[index];
-    return nil;
+    } else if (index == count) {
+        return NSLocalizedString(@"DOWNLOAD_SUBS_FROM_OSO", nil);
+    }
+    return @"";
 }
 
 - (NSString *)audioTrackNameAtIndex:(NSInteger)index
 {
     if (index >= 0 && index < _mediaPlayer.audioTrackNames.count)
         return _mediaPlayer.audioTrackNames[index];
-    return nil;
+    return @"";
 }
 
 - (NSDictionary *)titleDescriptionsDictAtIndex:(NSInteger)index
 {
     if (index >= 0 && index < _mediaPlayer.titleDescriptions.count)
         return _mediaPlayer.titleDescriptions[index];
-    return nil;
+    return [NSDictionary dictionary];
 }
 
 - (NSDictionary *)chapterDescriptionsDictAtIndex:(NSInteger)index
@@ -585,7 +636,7 @@ NSString *const VLCPlaybackServicePlaybackPositionUpdated = @"VLCPlaybackService
     NSArray *chapterDescriptions = [_mediaPlayer chapterDescriptionsOfTitle:_mediaPlayer.currentTitleIndex];
     if (index >= 0 && index < chapterDescriptions.count)
         return chapterDescriptions[index];
-    return nil;
+    return [NSDictionary dictionary];
 }
 
 - (void)selectAudioTrackAtIndex:(NSInteger)index
@@ -725,6 +776,15 @@ NSString *const VLCPlaybackServicePlaybackPositionUpdated = @"VLCPlaybackService
     [[NSNotificationCenter defaultCenter] postNotificationName:VLCPlaybackServicePlaybackDidPause object:self];
 }
 
+- (void)playItemAtIndex:(NSUInteger)index
+{
+    VLCMedia *media = [_mediaList mediaAtIndex:index];
+    [_listPlayer playItemAtNumber:[NSNumber numberWithUnsignedInteger:index]];
+    _mediaPlayer.media = media;
+    if ([self.delegate respondsToSelector:@selector(prepareForMediaPlayback:)])
+        [self.delegate prepareForMediaPlayback:self];
+}
+
 - (void)setShuffleMode:(BOOL)shuffleMode
 {
     _shuffleMode = shuffleMode;
@@ -781,12 +841,18 @@ NSString *const VLCPlaybackServicePlaybackPositionUpdated = @"VLCPlaybackService
 
     if (nextIndex < 0) {
         if (self.repeatMode == VLCRepeatAllItems) {
+#if TARGET_OS_IOS
+            [_delegate savePlaybackState:self];
+#endif
             [_listPlayer next];
             [[NSNotificationCenter defaultCenter]
              postNotificationName:VLCPlaybackServicePlaybackMetadataDidChange object:self];
         }
         return NO;
     }
+#if TARGET_OS_IOS
+    [_delegate savePlaybackState:self];
+#endif
 
     [_listPlayer playItemAtNumber:@(nextIndex)];
     [[NSNotificationCenter defaultCenter] postNotificationName:VLCPlaybackServicePlaybackMetadataDidChange object:self];
@@ -796,17 +862,23 @@ NSString *const VLCPlaybackServicePlaybackPositionUpdated = @"VLCPlaybackService
 - (BOOL)previous
 {
     if (_mediaList.count > 1) {
-        [_listPlayer previous];
-        [[NSNotificationCenter defaultCenter] postNotificationName:VLCPlaybackServicePlaybackMetadataDidChange object:self];
-    }
-    else {
+        VLCTime *playedTime = self.playedTime;
+        if (playedTime.value.longLongValue / 2000 >= 1) {
+            self.playbackPosition = .0;
+        } else {
+#if TARGET_OS_IOS
+            [_delegate savePlaybackState:self];
+#endif
+            [_listPlayer previous];
+            [[NSNotificationCenter defaultCenter] postNotificationName:VLCPlaybackServicePlaybackMetadataDidChange object:self];
+        }
+    } else {
         NSNumber *skipLength = [[NSUserDefaults standardUserDefaults] valueForKey:kVLCSettingPlaybackBackwardSkipLength];
         [_mediaPlayer jumpBackward:skipLength.intValue];
     }
     return YES;
 }
 
-
 - (void)jumpForward:(int)interval
 {
     [_mediaPlayer jumpForward:interval];
@@ -1051,6 +1123,16 @@ NSString *const VLCPlaybackServicePlaybackPositionUpdated = @"VLCPlaybackService
     return [_mediaPlayer preAmplification];
 }
 
+- (unsigned int)numberOfBands
+{
+    return [_mediaPlayer numberOfBands];
+}
+
+- (CGFloat)frequencyOfBandAtIndex:(unsigned int)index
+{
+    return [_mediaPlayer frequencyOfBandAtIndex:index];
+}
+
 #pragma mark - AVAudioSession Notification Observers
 
 - (void)handleInterruption:(NSNotification *)notification
@@ -1180,6 +1262,46 @@ NSString *const VLCPlaybackServicePlaybackPositionUpdated = @"VLCPlaybackService
     }
     [self restoreAudioAndSubtitleTrack];
 }
+
+- (void)_findCachedSubtitlesForMedia:(VLCMedia *)media
+{
+    /* if we already enforce a subtitle e.g. through Google Drive, don't try to find another */
+    if (_pathToExternalSubtitlesFile) {
+        return;
+    }
+    NSURL *mediaURL = media.url;
+    if (mediaURL.isFileURL) {
+        /* let's see if it is in the Inbox folder or outside our Documents folder and if yes, maybe we have a cached subtitles file? */
+        NSArray *searchPaths = NSSearchPathForDirectoriesInDomains(NSDocumentDirectory, NSUserDomainMask, YES);
+        NSString *documentFolderPath = [searchPaths firstObject];
+        NSString *potentialInboxFolderPath = [documentFolderPath stringByAppendingPathComponent:@"Inbox"];
+        NSFileManager *fileManager = [NSFileManager defaultManager];
+        NSString *mediaURLpath = mediaURL.path;
+        if ([mediaURLpath containsString:potentialInboxFolderPath] && ![mediaURLpath containsString:documentFolderPath]) {
+            NSString *mediaFileName = mediaURL.path.lastPathComponent.stringByDeletingPathExtension;
+            searchPaths = NSSearchPathForDirectoriesInDomains(NSCachesDirectory, NSUserDomainMask, YES);
+            NSString *cachefolderPath = [searchPaths.firstObject stringByAppendingPathComponent:kVLCSubtitlesCacheFolderName];
+
+            NSDirectoryEnumerator *folderEnumerator = [fileManager enumeratorAtURL:[NSURL fileURLWithPath:cachefolderPath]
+                                                        includingPropertiesForKeys:@[NSURLNameKey]
+                                                                           options:NSDirectoryEnumerationSkipsSubdirectoryDescendants
+                                                                      errorHandler:nil];
+            NSString *theSubtitleFileName;
+            for (NSURL *theURL in folderEnumerator) {
+                NSString *iter;
+                [theURL getResourceValue:&iter forKey:NSURLNameKey error:NULL];
+
+                if ([iter hasPrefix:mediaFileName]) {
+                    theSubtitleFileName = iter;
+                    break;
+                }
+            }
+
+            NSURL *subtitleURL = [NSURL fileURLWithPath:[cachefolderPath stringByAppendingPathComponent:theSubtitleFileName]];
+            [_mediaPlayer addPlaybackSlave:subtitleURL type:VLCMediaPlaybackSlaveTypeSubtitle enforce:YES];
+        }
+    }
+}
 #endif
 
 - (void)recoverDisplayedMetadata
@@ -1211,7 +1333,11 @@ NSString *const VLCPlaybackServicePlaybackPositionUpdated = @"VLCPlaybackService
 
 - (BOOL)isPlayingOnExternalScreen
 {
+#if !TARGET_OS_TV
     return (_renderer || [[UIDevice currentDevice] VLCHasExternalDisplay]);
+#else
+    return [[UIDevice currentDevice] VLCHasExternalDisplay];
+#endif
 }
 
 #pragma mark - background interaction
@@ -1234,12 +1360,17 @@ NSString *const VLCPlaybackServicePlaybackPositionUpdated = @"VLCPlaybackService
 {
     _preBackgroundWrapperView = _videoOutputViewWrapper;
 
+#if !TARGET_OS_TV
     if (!_renderer && _mediaPlayer.audioTrackIndexes.count > 0 && [_mediaPlayer isPlaying])
         [self setVideoTrackEnabled:false];
 
     if (_renderer) {
         [_backgroundDummyPlayer play];
     }
+#else
+    if (_mediaPlayer.audioTrackIndexes.count > 0 && [_mediaPlayer isPlaying])
+        [self setVideoTrackEnabled:false];
+#endif
 }
 
 - (void)applicationWillEnterForeground:(NSNotification *)notification
@@ -1249,9 +1380,11 @@ NSString *const VLCPlaybackServicePlaybackPositionUpdated = @"VLCPlaybackService
         _preBackgroundWrapperView = nil;
     }
 
+#if !TARGET_OS_TV
     if (_renderer) {
         [_backgroundDummyPlayer stop];
     }
+#endif
 
     if (_mediaPlayer.currentVideoTrackIndex == -1) {
         [self setVideoTrackEnabled:true];
@@ -1319,6 +1452,7 @@ NSString *const VLCPlaybackServicePlaybackPositionUpdated = @"VLCPlaybackService
 {
     float positionDiff = playbackTime - [self.metadata.elapsedPlaybackTime floatValue];
     [_mediaPlayer jumpForward:positionDiff];
+    _majorPositionChangeInProgress = 1;
 }
 
 #pragma mark - helpers
@@ -1331,17 +1465,21 @@ NSString *const VLCPlaybackServicePlaybackPositionUpdated = @"VLCPlaybackService
               kVLCSettingTextEncoding : [defaults objectForKey:kVLCSettingTextEncoding],
               kVLCSettingSkipLoopFilter : [defaults objectForKey:kVLCSettingSkipLoopFilter],
               kVLCSettingHardwareDecoding : [defaults objectForKey:kVLCSettingHardwareDecoding],
-              kVLCForceSMBV1 : [NSNull null]
+              kVLCSettingNetworkRTSPTCP : [defaults objectForKey:kVLCSettingNetworkRTSPTCP]
     };
 }
 
 #pragma mark - Renderer
+
+#if !TARGET_OS_TV
+
 - (void)setRenderer:(VLCRendererItem * __nullable)renderer
 {
     _renderer = renderer;
     [_mediaPlayer setRendererItem:_renderer];
 }
 
+#endif
 
 #pragma mark - PlayerDisplayController
 
@@ -1356,4 +1494,17 @@ NSString *const VLCPlaybackServicePlaybackPositionUpdated = @"VLCPlaybackService
     [_playerDisplayController dismissPlaybackView];
 }
 
+#pragma mark - VLCMediaListPlayerDelegate
+
+- (void)mediaListPlayer:(VLCMediaListPlayer *)player nextMedia:(VLCMedia *)media
+{
+#if TARGET_OS_IOS
+    [self _findCachedSubtitlesForMedia:media];
+#endif
+
+    if ([_delegate respondsToSelector:@selector(playbackService:nextMedia:)]) {
+        [_delegate playbackService:self nextMedia:media];
+    }
+}
+
 @end
